# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateEMDI(Template):
    'Fits 2-level random intercept Normal model using emdi package in R to a sample dataset and uses this with a population dataset to create SAE estimates.'

    __version__ = '1.0.0'

    tags = [ 'Model', '2-Level', 'SAE' ]
    engines = ['R_script']

    inputs = '''
estmeth = Text('Estimation Method: ', ['Model based', 'Direct'])
if estmeth == 'Model based':
    popdata = DataFileChooser('Population dataset: ')
y = DataVector('Response variable: ')
transform = Text('Transformation: ', ['box.cox', 'no', 'log'])
idcol = IDVector('Common 2 ID: ')
if estmeth == 'Model based':
     x = DataMatrix('Common predictor variables: ', allow_cat = True)
povind = Boolean('Do you want to calculate poverty related estimates, e.g. Head count ratio?: ')
if povind:
    userthresh = Text('Threshold value: ', ['Automatic', 'Manual'])
    if userthresh == 'Manual':
        thresh = Text('Threshold value')
    else:
        thresh = Text(value='')
ineqind = Boolean('Do you want to calculate inequality related estimates, e.g. GINI index?: ')
nproc = Integer('Parallel cores for bootstrapping: ', min=1, default=6)
if estmeth == 'Model based':
    # Note need to ensure pred is filled in before poppred line
    z = list(x)
    poppred = DataMatrix(value=list(x.orignames), categorical=x.categorical, data_src=popdata, allow_cat = True)
'''

    rscript = '''
% if estmeth == 'Model based':
popdata <- read.dta("${popdata}.dta")
% endif

# Tell R which encoding is used for the id variables
if (is.factor(mydata$${idcol})) {
  Encoding(levels(mydata$${idcol})) <- "UTF-8"
  #levels(mydata$${idcol}) <- enc2native(levels(mydata$${idcol}))
}
if (is.factor(popdata$${idcol})) {
  Encoding(levels(popdata$${idcol})) <- "UTF-8"
  #levels(popdata$${idcol}) <- enc2native(levels(popdata$${idcol}))
}

local({r <- getOption("repos"); r["CRAN"] <- "http://cran.r-project.org"; options(repos = r)})
if (!require(emdi)) {
    install.packages("emdi")
    library(emdi)
}

if (!require(tableHTML)) {
    install.packages("tableHTML")
    library(tableHTML)
}

my_indicators <- list(stdev = function(y, threshold) { sd(y) })

threshold = NULL
% if povind:
% if userthresh == 'Manual':
threshold = ${thresh}
% endif
% if userthresh == 'Automatic':
threshold = function(y){0.6 * median(y)}
% endif
% endif

% if estmeth == 'Model based':
mymodel <- ebp(fixed = ${y} ~ ${'+'.join(x)},
               transformation="${transform}",
               pop_data = popdata,
               pop_domains = "${idcol}",
               smp_data = mydata,
               smp_domains = "${idcol}",
               threshold = threshold,
               MSE=TRUE,
               cpus=${nproc},
               custom_indicator=my_indicators
              )
% endif
% if estmeth == 'Direct':
mymodel <- direct(y = "${y}",
                  transformation=${transform},
                  smp_data = mydata, smp_domains = "${idcol}",
                  custom_indicator=my_indicators
                 )
% endif

summary(mymodel)

summ <- summary(mymodel)

est_mean = estimators(mymodel, MSE=TRUE, indicator="Mean")
est_quant = estimators(mymodel, indicator="Quantiles")
est_pov =  estimators(mymodel, MSE=TRUE, indicator="Poverty")
est_ineq = estimators(mymodel, MSE=TRUE, indicator="Inequality")
est_cust = estimators(mymodel, indicator="Custom")
estimates <- merge(est_mean, est_quant, by="Domain")
% if povind:
estimates <- merge(estimates, est_pov, by="Domain")
% endif
% if ineqind:
estimates <- merge(estimates, est_ineq, by="Domain")
% endif
estimates <- merge(estimates, est_cust, by="Domain")

# Sort Estimates
estimates <- estimates[order(as.integer(estimates$Domain)), ]

write.dta(estimates, "estimates.dta")
write_tableHTML(tableHTML(estimates, round=2), file="estimates.html")

sink("sae_prediction_text.html")
cat("<p>")
cat("Our primary interest is in the small area estimates that this model produces.")
cat("The table below gives estimates for the mean and SD for each of the ", length(unique(estimates$Domain)), " small areas in the dataset.")
cat("These estimates can be found in the columns headed population whilst for comparison in the columns headed sample are the means and SDS for ${y} just using the sample data.")
cat("</p>")
sink()

smpmean <- aggregate(${y} ~ ${idcol}, data = mydata, FUN = mean)
colnames(smpmean) <- c("${idcol}", "Sample_mean")

smpsd <- aggregate(${y} ~ ${idcol}, data = mydata, FUN = sd)
colnames(smpsd) <- c("${idcol}", "Sample_sd")

prediction_summary <- data.frame(${idcol} = estimates$Domain, Population_mean = estimates$Mean, Population_sd = estimates$stdev)
prediction_summary <- merge(prediction_summary, smpmean, by="${idcol}", all = TRUE)
prediction_summary <- merge(prediction_summary, smpsd, by="${idcol}", all = TRUE)
prediction_summary <- prediction_summary[c("${idcol}", "Sample_mean", "Sample_sd", "Population_mean", "Population_sd")]

write.dta(prediction_summary, "prediction_summary.dta")
write_tableHTML(tableHTML(prediction_summary, round=2), file="prediction_summary.html")

write.dta(data.frame(summ$size_dom), "size_dom.dta")
write_tableHTML(tableHTML(summ$size_dom, round=2), file="size_dom.html")

if (!is.null(summ$transform)) {
    write.dta(summ$transform, "transform.dta")
    write_tableHTML(tableHTML(summ$transform, round=2), file="transform.html")
}
if (!is.null(summ$normality)) {
    write.dta(summ$normality, "normality.dta")
    write_tableHTML(tableHTML(summ$normality, round=2), file="normality.html")
}
if (!is.null(summ$coeff_determ)) {
    write.dta(summ$coeff_determ, "coeff_determ.dta")
    write_tableHTML(tableHTML(summ$normality, round=2), file="normality.html")
}
if (!is.null(summ$size_dom)) {
    write.dta(data.frame(summ$size_dom), "size_dom.dta")
    write_tableHTML(tableHTML(summ$size_dom, round=2), file="size_dom.html")
}
if (!is.null(summ$smp_size_tab)) {
    write.dta(data.frame(summ$smp_size_tab), "smp_size_tab.dta")
    write_tableHTML(tableHTML(summ$smp_size_tab, round=2), file="smp_size_tab.html")
}
% if estmeth == 'Model based':
sample <- data.frame(out_of_smp=summ$out_of_smp, in_smp=summ$in_smp, size_smp=summ$size_smp, size_pop=summ$size_pop)
write.dta(sample, "sample.dta")
write_tableHTML(tableHTML(sample, round=2), file="sample.html")
% endif
% if estmeth == 'Direct':
sample <- data.frame(in_smp=summ$in_smp, size_smp=summ$size_smp)
write.dta(sample, "sample.dta")
write_tableHTML(tableHTML(sample, round=2), file="sample.html")
% endif
if (!is.null(summ$icc)) {
    icc <- data.frame(icc=summ$icc)
    write.dta(icc, "icc.dta")
    write_tableHTML(tableHTML(icc, round=2), file="icc.html")
}

% if estmeth == 'Model based':
svg()
plot(mymodel)
dev.off()
% endif

caterpillar <- function(y, x, qtlow, qtup, xlab = "", ylab = "", xlim = NULL, ylim = NULL, main = "") {
  y[!is.finite(y)] <- 0
  qtlow[!is.finite(qtlow)] <- 0
  qtup[!is.finite(qtup)] <- 0
  if (is.null(ylim)){
    ylim = c(min(qtlow), max(qtup))
  }
  plot(x, y, xlim = xlim, ylim = ylim, pch = 15, xlab = xlab, ylab = ylab, main = main)
  points(x, qtlow, pch = 24, bg = "grey")
  points(x, qtup, pch = 25, bg = "grey")
  for (i in 1:length(x)) {
    lines(rep(x[i], 2), c(qtlow[i], qtup[i]))
  }
}

svg("sae_mean.svg")
meanerr <- sqrt(estimates$Mean_MSE)
caterpillar(x = prediction_summary$${idcol},
            y = prediction_summary$Population_mean,
            qtlow = prediction_summary$Population_mean-1.96*meanerr,
            qtup = prediction_summary$Population_mean+1.96*meanerr,
            xlab = "${idcol}",
            ylab = "${y}",
            main = "Prediction for ${y} (mean)"
           )
dev.off()

quantiles_table <- data.frame(${idcol} = estimates$Domain,
                              q10 = estimates$Quantile_10,
                              q25 = estimates$Quantile_25,
                              q50 = estimates$Median,
                              q75 = estimates$Quantile_75,
                              q90 = estimates$Quantile_90
                             )
write.dta(quantiles_table, "sae_quantilestable.dta")
write_tableHTML(tableHTML(quantiles_table, round=2), file="sae_quantilestable.html")

#origmar = par("mar")
svg("sae_quantiles.svg")
par(mar = c(5, 4, 4, 8))
colours <- rainbow(5)
plot(estimates$Domain, estimates$Median, type="n", xlab="${idcol}", ylab="${y}", ylim = c(min(estimates$Quantile_10), max(estimates$Quantile_90)))
lines(estimates$Domain, estimates$Quantile_10, col=colours[1])
lines(estimates$Domain, estimates$Quantile_25, col=colours[2])
lines(estimates$Domain, estimates$Median, col=colours[3])
lines(estimates$Domain, estimates$Quantile_75, col=colours[4])
lines(estimates$Domain, estimates$Quantile_90, col=colours[5])
title("Prediction for ${y} (quantiles)")
legend(x = 'right', legend = c("10%", "25%", "50%", "75%", "90%"), col = colours, lty = 1, title = "Quantiles", inset = c(-0.3, -0.3), xpd = TRUE)
dev.off()
#par(mar = origmar)

% if ineqind:
sink("sae_ginitext.html")
cat("<p>")
cat("Below is a graph of the Gini index (with confidence intervals calculated by bootstrapping) for the ", length(unique(estimates$Domain)), " small areas. ")
cat("</p>")
sink()

svg("sae_gini.svg")
ginierr <- sqrt(estimates$Gini_MSE)
caterpillar(x=estimates$Domain,
            y=estimates$Gini,
            qtlow=estimates$Gini-1.96*ginierr,
            qtup=estimates$Gini+1.96*ginierr,
            xlab = "${idcol}",
            ylab = "GINI",
            main = "Prediction for ${y} (GINI)"
           )
dev.off()

sink("sae_qsharedesc.html")
cat("<p>")
cat("The quintile share ratio (QSR) measures the ratio of the total of the response (usually income) received by the top 20% of the population to the total of the response received by the bottom 20% of the population. ")
cat("So a value of 2 would mean that on average individuals in the top 20% would have values of ${y} twice as high as individuals in the bottom 20%. ")
cat("</p>")
sink()

sink("sae_qsharetext.html")
cat("<p>")
cat("Below is a graph of the quintile share ratio (with confidence intervals calculated by bootstrapping) for the ", length(unique(estimates$Domain)), " small areas. ")
cat("</p>")
sink()

svg("sae_qshare.svg")
qshareerr <- sqrt(estimates$Quintile_Share_MSE)
caterpillar(x = estimates$Domain,
            y = estimates$Quintile_Share,
            qtlow = estimates$Quintile_Share-1.96*qshareerr,
            qtup = estimates$Quintile_Share+1.96*qshareerr,
            xlab = "${idcol}",
            ylab = "Quintile Share",
            main = "Prediction for ${y} (quintile share)"
           )
dev.off()
% endif

% if povind:

sink("sae_headcountdesc.html")
cat("<p>")
cat("Next we have the Head count ratio (HCR). ")
cat("This is again often associated with income and poverty and represents the proportion of each small area whose value of ${y} lies below a pre-specified threshold. ")
cat("</p>")
sink()

sink("sae_headcounttext.html")
cat("<p>")
cat("Below is a graph of the head count ratio (with confidence intervals calculated by bootstrapping) for the ", length(unique(estimates$Domain)), " small areas. ")
cat("</p>")
sink()

svg("sae_headcount.svg")
hcounterr <- sqrt(estimates$Head_Count_MSE)
caterpillar(x = estimates$Domain,
            y = estimates$Head_Count,
            qtlow = estimates$Head_Count-1.96*hcounterr,
            qtup = estimates$Head_Count+1.96*hcounterr,
            xlab = "${idcol}",
            ylab = "Head Count",
            main = "Prediction for ${y} (head count)"
           )
dev.off()

sink("sae_povgaptext.html")
cat("<p>")
cat("Below is a graph of the poverty gap index (with confidence intervals calculated by bootstrapping) for the ", length(unique(estimates$Domain)), " small areas. ")
cat("</p>")
sink()

svg("sae_povgap.svg")
povgaperr <- sqrt(estimates$Poverty_Gap_MSE)
caterpillar(x = estimates$Domain,
            y = estimates$Poverty_Gap,
            qtlow = estimates$Poverty_Gap-1.96*povgaperr,
            qtup = estimates$Poverty_Gap+1.96*povgaperr,
            xlab = "${idcol}",
            ylab = "Poverty Gap",
            main = "Prediction for ${y} (poverty gap)"
           )
dev.off()
% endif

% if ineqind or povind:

sae_summary <- data.frame(${idcol} = estimates$Domain)
% if ineqind:
sae_summary <- merge(sae_summary, data.frame(${idcol} = estimates$Domain, Gini = estimates$Gini, Gini_Rank = rank(estimates$Gini), QSR = estimates$Quintile_Share, QSR_Rank = rank(estimates$Quintile_Share)), by="${idcol}")
% endif
% if povind:
sae_summary <- merge(sae_summary, data.frame(${idcol} = estimates$Domain, HCR = estimates$Head_Count, HCR_Rank = rank(estimates$Head_Count), PGI = estimates$Poverty_Gap, PGI_Rank = rank(estimates$Poverty_Gap)), by="${idcol}")
% endif

# Sort Summary
sae_summary <- sae_summary[order(as.integer(sae_summary$${idcol})), ]
rownames(sae_summary) <- 1:nrow(sae_summary)

write.dta(sae_summary, "sae_summarystats.dta")
write_tableHTML(tableHTML(sae_summary, round=2), file="sae_summarystats.html")

% endif
'''
