# Copyright (c) 2017, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateSAEResponse(Template):
    'Explores the response variable for SAE models and is used in the SAE eBook.'

    __version__ = '1.0.0'

    tags = [ 'Data manipulation', 'SAE' ]
    engines = ['Python_script']

    inputs = '''
popdata = DataFileChooser('Population dataset: ')
resp = DataVector('Response variable: ')
idcol = IDVector('Common ID variable: ')
lambval = Text('Lambda parameter for response tranformation: ')
'''


    pythonscript = '''
import numpy

from io import BytesIO

from matplotlib.figure import Figure
import matplotlib.lines as lines
from matplotlib.backends.backend_agg import FigureCanvasAgg

import sys
import os
sys.path.append(os.path.abspath(os.path.join('..', '..', 'apps', 'cmdtest')))
from cmdtest import RunStatJR

import EStat
from EStat.Templating import *
from EStat.DTAFile import DTAFile


def roundtosf(num, sf):
    import numpy
    return numpy.round(num, sf-int(numpy.trunc(numpy.log10(numpy.abs(num)))))

data2 = globals()[str(popdata)]

datafile.addvariable("Intercept", data=numpy.ones(datafile.nobs))
data2.addvariable("Intercept", data=numpy.ones(data2.nobs))

datasets.datasets['workdata'] = datafile
datasets.datasets['workdata2'] = data2

sampvars = []
sampvars.append(datafile.variables[str(idcol)]['data'])
sampvars.append(numpy.ones(datafile.nobs))

popvars = []
popvars.append(data2.variables[str(idcol)]['data'])
popvars.append(numpy.zeros(data2.nobs))

print("Sample:")
print(numpy.array(sampvars))

print("Population")
print(numpy.array(popvars))

allvars = numpy.concatenate((sampvars, popvars), axis=1)

print("Joined:")
print(allvars)

retval = DTAFile()
retval.nobs = allvars.shape[1]
retval.addvariable(str(idcol), data = numpy.array(allvars[-2], dtype=int))
retval.variables[str(idcol)]['varlabel'] = datafile.variables[str(idcol)]['varlabel']
lab = datafile.variables[str(idcol)]['label']
retval.variables[str(idcol)]['label'] = lab
if lab in datafile.valuelabels.keys():
    retval.valuelabels[lab] = datafile.valuelabels[lab]
if lab in data2.valuelabels.keys():
    retval.valuelabels[lab].update(data2.valuelabels[lab])

retval.addvariable('_sample_ind', data = numpy.array(allvars[-1], dtype=int))

outputs['result'] = retval

uniqids = numpy.unique(retval.variables[str(idcol)]['data'])

uniqpopids = numpy.unique(data2.variables[str(idcol)]['data'])
uniqsampids = numpy.unique(datafile.variables[str(idcol)]['data'])

html = '<p>'
html += 'In your dataset the small areas are represented by the variable name ' + str(idcol) + ' and there are ' + str(data2.nobs) + ' individuals in the population that come from ' + str(len(uniqpopids)) + ' areas in total. '
html += 'The sample dataset has ' + str(datafile.nobs) + ' individuals in total (' + str(round(float(datafile.nobs) / float(data2.nobs) * 100.0, 2)) + '% of the population) with individuals in the sample coming from ' + str(len(uniqsampids)) + ' areas.'
html += '</p>'

outputs['datadist.html'] = HTMLOutput(html, "Distribution of data in the population and sample data sets")

tabout = TabularOutput()
tabout.column_headings = ['Code', 'Name', 'Nsamp', 'Npop', '%']

for id in uniqids:
    if lab in retval.valuelabels and id in retval.valuelabels[lab]:
        name = retval.valuelabels[lab][id]
    else:
        name = str(idcol) + '==' + str(id)
    npop = numpy.sum((retval.variables[str(idcol)]['data'] == id) & (retval.variables['_sample_ind']['data'] == 0))
    nsamp = numpy.sum((retval.variables[str(idcol)]['data'] == id) & (retval.variables['_sample_ind']['data'] == 1))
    tabout.add_row(str(id), [name, nsamp, npop, float(nsamp)/float(npop)*100.0])

outputs['coverage'] = tabout

# VPC

m = RunStatJR(template='2LevelMod', dataset='workdata', invars = {'y': str(resp), 'L2ID': str(idcol), 'storeresid': 'No', 'D': 'Normal', 'x': 'Intercept'}, estoptions = {'Engine': 'R_lme4', 'defaultalg': 'Yes', 'nAGQ': '1'})
var2 = m.eng.outputs['ModelResults'].content['paramstats']['est'][str(idcol) + '__Intercept_']
var1 = m.eng.outputs['ModelResults'].content['paramstats']['est']['Residual']
vpc = var2 / (var2 + var1)

html = '<p>'
html += 'When we look at ' + str(resp) + ' in the sample dataset there is variability in the average response across the small areas that we are estimating values for. '
html += 'It is important to estimate how much variation in the response is between small areas and how much is within small areas and this is done with a statistic called the VPC. '
html += 'Here the VPC is ' + str(numpy.round(vpc, 2)) + '. '
html += 'This means that ' + str(int(numpy.round(vpc * 100, 0))) + '% of the variation in ' + str(resp) + ' is between small areas and therefore due to differences across areas. '
html += 'These differences may be explained by the predictor variables in our later modelling. '
html += '</p>'

outputs['vpctext.html'] = HTMLOutput(html, "VPC text")


# Histogram

respdata = datafile.variables[str(resp)]['data']

fig = Figure(figsize=(4,4))
ax = fig.add_subplot(100 + 10 + 1)
ax.set_xlabel(str(resp), fontsize=7)
ax.set_ylabel('Frequency', fontsize=7)
for label in (ax.get_xticklabels() + ax.get_yticklabels()):
    label.set_fontsize(5)
ax.hist(respdata, 20)
normx = numpy.linspace(respdata.min(), respdata.max(), 100)
normy = scipy.stats.norm.pdf(normx, respdata.mean(), respdata.std()) * len(respdata) * (respdata.max() - respdata.min()) / 20.0
ax.plot(normx, normy, linewidth=3)

canvas = FigureCanvasAgg(fig)
buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['histogram.svg'] = ImageOutput(buf.getvalue())
buf.close()

# Skew text
skew = scipy.stats.skew(respdata)
skewtest = scipy.stats.skewtest(respdata)
html = '<p>'
if skewtest[1] > 0.05:
    html += 'Here the distribution is reasonably symmetric with skewness value ' + str(numpy.round(skew,3)) + '. '
else:
    if skew > 0:
        html += 'Here the median is smaller than the mean and there is significant skew to the right. '
        if skew < 2:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + '. '
            html += 'Here the statistical significance may be to some degree due to the large sample size as from a practical perspective values of skew less than 2 are not considered too big a skew. '
        else:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + ' and so this is practically large as well as statistically significant. '
    else:
        html += 'Here the median is larger than the mean and there is significant skew to the left. '
        if skew > -2:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + '. '
            html += 'Here the statistical significance may be to some degree due to the large sample size as from a practical perspective values of skew less than 2 in magnitude are not considered too big a skew. '
        else:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + ' and so this is practically large as well as statistically significant. '
html += '</p>'

outputs['skewtext.html'] = HTMLOutput(html, "Skewness")

logresp = numpy.log(respdata)

fig = Figure(figsize=(4,4))
ax = fig.add_subplot(100 + 10 + 1)
ax.set_xlabel('Log(' + str(resp) + ')', fontsize=7)
ax.set_ylabel('Frequency', fontsize=7)
for label in (ax.get_xticklabels() + ax.get_yticklabels()):
    label.set_fontsize(5)
ax.hist(logresp, 20)
normx = numpy.linspace(logresp.min(), logresp.max(), 100)
normy = scipy.stats.norm.pdf(normx, logresp.mean(), logresp.std()) * len(logresp) * (logresp.max() - logresp.min()) / 20.0
ax.plot(normx, normy, linewidth=3)

canvas = FigureCanvasAgg(fig)
buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['loghistogram.svg'] = ImageOutput(buf.getvalue())
buf.close()

# Skew text
skew = scipy.stats.skew(logresp)
skewtest = scipy.stats.skewtest(logresp)
html = '<p>'
if skewtest[1] > 0.05:
    html += 'Here the distribution is reasonably symmetric with skewness value ' + str(numpy.round(skew,3)) + '. '
else:
    if skew > 0:
        html += 'Here the median is smaller than the mean and there is significant skew to the right. '
        if skew < 2:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + '. '
            html += 'Here the statistical significance may be to some degree due to the large sample size as from a practical perspective values of skew less than 2 are not considered too big a skew. '
        else:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + ' and so this is practically large as well as statistically significant. '
    else:
        html += 'Here the median is larger than the mean and there is significant skew to the left. '
        if skew > -2:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + '. '
            html += 'Here the statistical significance may be to some degree due to the large sample size as from a practical perspective values of skew less than 2 in magnitude are not considered too big a skew. '
        else:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + ' and so this is practically large as well as statistically significant. '
html += '</p>'

outputs['logskewtext.html'] = HTMLOutput(html, "Skewness")

bcresp = (respdata**float(lambval) - 1) / float(lambval)

fig = Figure(figsize=(4,4))
ax = fig.add_subplot(100 + 10 + 1)
ax.set_xlabel('Box-Cox(' + str(resp) + ')', fontsize=7)
ax.set_ylabel('Frequency', fontsize=7)
for label in (ax.get_xticklabels() + ax.get_yticklabels()):
    label.set_fontsize(5)
ax.hist(bcresp, 20)
normx = numpy.linspace(bcresp.min(), bcresp.max(), 100)
normy = scipy.stats.norm.pdf(normx, bcresp.mean(), bcresp.std()) * len(bcresp) * (bcresp.max() - bcresp.min()) / 20.0
ax.plot(normx, normy, linewidth=3)

canvas = FigureCanvasAgg(fig)
buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['bchistogram.svg'] = ImageOutput(buf.getvalue())
buf.close()

# Skew text
skew = scipy.stats.skew(bcresp)
skewtest = scipy.stats.skewtest(bcresp)
html = '<p>'
if skewtest[1] > 0.05:
    html += 'Here the distribution is reasonably symmetric with skewness value ' + str(numpy.round(skew,3)) + '. '
else:
    if skew > 0:
        html += 'Here the median is smaller than the mean and there is significant skew to the right. '
        if skew < 2:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + '. '
            html += 'Here the statistical significance may be to some degree due to the large sample size as from a practical perspective values of skew less than 2 are not considered too big a skew. '
        else:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + ' and so this is practically large as well as statistically significant. '
    else:
        html += 'Here the median is larger than the mean and there is significant skew to the left. '
        if skew > -2:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + '. '
            html += 'Here the statistical significance may be to some degree due to the large sample size as from a practical perspective values of skew less than 2 in magnitude are not considered too big a skew. '
        else:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + ' and so this is practically large as well as statistically significant. '
html += '</p>'

outputs['bcskewtext.html'] = HTMLOutput(html, "Skewness")

dpresp = (respdata**float(lambval) - respdata**-float(lambval)) / 2 * float(lambval)

fig = Figure(figsize=(4,4))
ax = fig.add_subplot(100 + 10 + 1)
ax.set_xlabel('Dual-Power(' + str(resp) + ')', fontsize=7)
ax.set_ylabel('Frequency', fontsize=7)
for label in (ax.get_xticklabels() + ax.get_yticklabels()):
    label.set_fontsize(5)
ax.hist(dpresp, 20)
normx = numpy.linspace(dpresp.min(), dpresp.max(), 100)
normy = scipy.stats.norm.pdf(normx, dpresp.mean(), dpresp.std()) * len(dpresp) * (dpresp.max() - dpresp.min()) / 20.0
ax.plot(normx, normy, linewidth=3)

canvas = FigureCanvasAgg(fig)
buf = BytesIO()
canvas.print_figure(buf, dpi=80, format='svg')

buf.seek(0)
outputs['dphistogram.svg'] = ImageOutput(buf.getvalue())
buf.close()

# Skew text
skew = scipy.stats.skew(dpresp)
skewtest = scipy.stats.skewtest(dpresp)
html = '<p>'
if skewtest[1] > 0.05:
    html += 'Here the distribution is reasonably symmetric with skewness value ' + str(numpy.round(skew,3)) + '. '
else:
    if skew > 0:
        html += 'Here the median is smaller than the mean and there is significant skew to the right. '
        if skew < 2:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + '. '
            html += 'Here the statistical significance may be to some degree due to the large sample size as from a practical perspective values of skew less than 2 are not considered too big a skew. '
        else:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + ' and so this is practically large as well as statistically significant. '
    else:
        html += 'Here the median is larger than the mean and there is significant skew to the left. '
        if skew > -2:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + '. '
            html += 'Here the statistical significance may be to some degree due to the large sample size as from a practical perspective values of skew less than 2 in magnitude are not considered too big a skew. '
        else:
            html += 'The skewness value is ' + str(numpy.round(skew, 3)) + ' and so this is practically large as well as statistically significant. '
html += '</p>'

outputs['dpskewtext.html'] = HTMLOutput(html, "Skewness")


# Outliers
#sdout = 0.8

#limithigh = respdata.mean() + (respdata.std() * float(sdout))
#limitlow = respdata.mean() - (respdata.std() * float(sdout))
#outpos = numpy.where(numpy.logical_or(respdata < limitlow, respdata > limithigh))[0]
#outliers = respdata[outpos]
#if len(outliers) > 0:
#    tabout = TabularOutput()
#    tabout.column_headings = ['Row', str(resp)]
#
#    for pos in outpos:
#        tabout.add_row(str(pos+1), [respdata[pos]])
#    outputs['outliers'] = tabout
'''